import phonemizer
import re
import torch
import numpy as np

def split_num(num):
    num = num.group()
    if '.' in num:
        return num
    elif ':' in num:
        h, m = [int(n) for n in num.split(':')]
        if m == 0:
            return f"{h} o'clock"
        elif m < 10:
            return f'{h} oh {m}'
        return f'{h} {m}'
    year = int(num[:4])
    if year < 1100 or year % 1000 < 10:
        return num
    left, right = num[:2], int(num[2:4])
    s = 's' if num.endswith('s') else ''
    if 100 <= year % 1000 <= 999:
        if right == 0:
            return f'{left} hundred{s}'
        elif right < 10:
            return f'{left} oh {right}{s}'
    return f'{left} {right}{s}'

def flip_money(m):
    m = m.group()
    bill = 'dollar' if m[0] == '$' else 'pound'
    if m[-1].isalpha():
        return f'{m[1:]} {bill}s'
    elif '.' not in m:
        s = '' if m[1:] == '1' else 's'
        return f'{m[1:]} {bill}{s}'
    b, c = m[1:].split('.')
    s = '' if b == '1' else 's'
    c = int(c.ljust(2, '0'))
    coins = f"cent{'' if c == 1 else 's'}" if m[0] == '$' else ('penny' if c == 1 else 'pence')
    return f'{b} {bill}{s} and {c} {coins}'

def point_num(num):
    a, b = num.group().split('.')
    return ' point '.join([a, ' '.join(b)])

def normalize_text(text):
    text = text.replace(chr(8216), "'").replace(chr(8217), "'")
    text = text.replace('«', chr(8220)).replace('»', chr(8221))
    text = text.replace(chr(8220), '"').replace(chr(8221), '"')
    text = text.replace('(', '«').replace(')', '»')
    for a, b in zip('、。！，：；？', ',.!,:;?'):
        text = text.replace(a, b+' ')
    text = re.sub(r'[^\S \n]', ' ', text)
    text = re.sub(r'  +', ' ', text)
    text = re.sub(r'(?<=\n) +(?=\n)', '', text)
    text = re.sub(r'\bD[Rr]\.(?= [A-Z])', 'Doctor', text)
    text = re.sub(r'\b(?:Mr\.|MR\.(?= [A-Z]))', 'Mister', text)
    text = re.sub(r'\b(?:Ms\.|MS\.(?= [A-Z]))', 'Miss', text)
    text = re.sub(r'\b(?:Mrs\.|MRS\.(?= [A-Z]))', 'Mrs', text)
    text = re.sub(r'\betc\.(?! [A-Z])', 'etc', text)
    text = re.sub(r'(?i)\b(y)eah?\b', r"\1e'a", text)
    text = re.sub(r'\d*\.\d+|\b\d{4}s?\b|(?<!:)\b(?:[1-9]|1[0-2]):[0-5]\d\b(?!:)', split_num, text)
    text = re.sub(r'(?<=\d),(?=\d)', '', text)
    text = re.sub(r'(?i)[$£]\d+(?:\.\d+)?(?: hundred| thousand| (?:[bm]|tr)illion)*\b|[$£]\d+\.\d\d?\b', flip_money, text)
    text = re.sub(r'\d*\.\d+', point_num, text)
    text = re.sub(r'(?<=\d)-(?=\d)', ' to ', text)
    text = re.sub(r'(?<=\d)S', ' S', text)
    text = re.sub(r"(?<=[BCDFGHJ-NP-TV-Z])'?s\b", "'S", text)
    text = re.sub(r"(?<=X')S\b", 's', text)
    text = re.sub(r'(?:[A-Za-z]\.){2,} [a-z]', lambda m: m.group().replace('.', '-'), text)
    text = re.sub(r'(?i)(?<=[A-Z])\.(?=[A-Z])', '-', text)
    return text.strip()

def get_vocab():
    _pad = "$"
    _punctuation = ';:,.!?¡¿—…"«»“” '
    _letters = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz'
    _letters_ipa = "ɑɐɒæɓʙβɔɕçɗɖðʤəɘɚɛɜɝɞɟʄɡɠɢʛɦɧħɥʜɨɪʝɭɬɫɮʟɱɯɰŋɳɲɴøɵɸθœɶʘɹɺɾɻʀʁɽʂʃʈʧʉʊʋⱱʌɣɤʍχʎʏʑʐʒʔʡʕʢǀǁǂǃˈˌːˑʼʴʰʱʲʷˠˤ˞↓↑→↗↘'̩'ᵻ"
    symbols = [_pad] + list(_punctuation) + list(_letters) + list(_letters_ipa)
    dicts = {}
    for i in range(len((symbols))):
        dicts[symbols[i]] = i
    return dicts

VOCAB = get_vocab()
def tokenize(ps):
    return [i for i in map(VOCAB.get, ps) if i is not None]

phonemizers = dict(
    a=phonemizer.backend.EspeakBackend(language='en-us', preserve_punctuation=True, with_stress=True),
    b=phonemizer.backend.EspeakBackend(language='en-gb', preserve_punctuation=True, with_stress=True),
)
def phonemize(text, lang, norm=True):
    if norm:
        text = normalize_text(text)
    ps = phonemizers[lang].phonemize([text])
    ps = ps[0] if ps else ''
    # https://en.wiktionary.org/wiki/kokoro#English
    ps = ps.replace('kəkˈoːɹoʊ', 'kˈoʊkəɹoʊ').replace('kəkˈɔːɹəʊ', 'kˈəʊkəɹəʊ')
    ps = ps.replace('ʲ', 'j').replace('r', 'ɹ').replace('x', 'k').replace('ɬ', 'l')
    ps = re.sub(r'(?<=[a-zɹː])(?=hˈʌndɹɪd)', ' ', ps)
    ps = re.sub(r' z(?=[;:,.!?¡¿—…"«»“” ]|$)', 'z', ps)
    if lang == 'a':
        ps = re.sub(r'(?<=nˈaɪn)ti(?!ː)', 'di', ps)
    ps = ''.join(filter(lambda p: p in VOCAB, ps))
    return ps.strip()

def length_to_mask(lengths):
    mask = torch.arange(lengths.max()).unsqueeze(0).expand(lengths.shape[0], -1).type_as(lengths)
    mask = torch.gt(mask+1, lengths.unsqueeze(1))
    return mask

@torch.no_grad()
def forward(model, tokens, ref_s, speed):
    device = ref_s.device
    tokens = torch.LongTensor([[0, *tokens, 0]]).to(device)
    input_lengths = torch.LongTensor([tokens.shape[-1]]).to(device)
    text_mask = length_to_mask(input_lengths).to(device)
    bert_dur = model.bert(tokens, attention_mask=(~text_mask).int())
    d_en = model.bert_encoder(bert_dur).transpose(-1, -2)
    s = ref_s[:, 128:]
    d = model.predictor.text_encoder(d_en, s, input_lengths, text_mask)
    x, _ = model.predictor.lstm(d)
    duration = model.predictor.duration_proj(x)
    duration = torch.sigmoid(duration).sum(axis=-1) / speed
    pred_dur = torch.round(duration).clamp(min=1).long()
    pred_aln_trg = torch.zeros(input_lengths, pred_dur.sum().item())
    c_frame = 0
    for i in range(pred_aln_trg.size(0)):
        pred_aln_trg[i, c_frame:c_frame + pred_dur[0,i].item()] = 1
        c_frame += pred_dur[0,i].item()
    en = d.transpose(-1, -2) @ pred_aln_trg.unsqueeze(0).to(device)
    F0_pred, N_pred = model.predictor.F0Ntrain(en, s)
    t_en = model.text_encoder(tokens, input_lengths, text_mask)
    asr = t_en @ pred_aln_trg.unsqueeze(0).to(device)
    return model.decoder(asr, F0_pred, N_pred, ref_s[:, :128]).squeeze().cpu().numpy()

def generate(model, text, voicepack, lang='a', speed=1, ps=None):
    ps = ps or phonemize(text, lang)
    tokens = tokenize(ps)
    if not tokens:
        return None
    elif len(tokens) > 510:
        tokens = tokens[:510]
        print('Truncated to 510 tokens')
    ref_s = voicepack[len(tokens)]
    out = forward(model, tokens, ref_s, speed)
    ps = ''.join(next(k for k, v in VOCAB.items() if i == v) for i in tokens)
    return out, ps

def generate_full(model, text, voicepack, lang='a', speed=1, ps=None):
    ps = ps or phonemize(text, lang)
    tokens = tokenize(ps)
    if not tokens:
        return None
    outs = []
    loop_count = len(tokens)//510 + (1 if len(tokens) % 510 != 0 else 0)
    for i in range(loop_count):
        ref_s = voicepack[len(tokens[i*510:(i+1)*510])]
        out = forward(model, tokens[i*510:(i+1)*510], ref_s, speed)
        outs.append(out)
    outs = np.concatenate(outs)
    ps = ''.join(next(k for k, v in VOCAB.items() if i == v) for i in tokens)
    return outs, ps


from verse.core import OperationParser


class TextEmbeddingOperationParser(OperationParser):
    def get_texts(self) -> list[str]:
        texts = self.get_arg("texts")
        if isinstance(texts, str):
            return [texts]
        return texts

    def get_embedding1(self) -> list[float]:
        return self.get_arg("embedding1")

    def get_embedding2(self) -> list[float]:
        return self.get_arg("embedding2")



from ._models import (
    ChatCompletionMessage,
    ChatCompletionMessageContentPart,
    ChatCompletionResult,
    ChatCompletionTokenLogprob,
    Choice,
    ChoiceLogProbs,
    FunctionDefinition,
    JSONSchema,
    ResponseFormat,
    Stream,
    StreamOptions,
    Tool,
    ToolCall,
    ToolCallFunction,
    ToolChoice,
    ToolChoiceFunction,
    TopLogprob,
    Usage,
)
from .component import ChatCompletion

__all__ = [
    "ChatCompletion",
    "ChatCompletionMessage",
    "ChatCompletionMessageContentPart",
    "ChatCompletionResult",
    "ChatCompletionTokenLogprob",
    "Choice",
    "ChoiceLogProbs",
    "FunctionDefinition",
    "JSONSchema",
    "ResponseFormat",
    "Stream",
    "StreamOptions",
    "Tool",
    "ToolCall",
    "ToolCallFunction",
    "ToolChoice",
    "ToolChoiceFunction",
    "TopLogprob",
    "Usage",
]



from typing import Any

from verse.content.audio import Audio, AudioParam
from verse.core import Component, Response, operation


class AudioEmbedding(Component):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @operation()
    def embed(
        self,
        audio: str | bytes | Audio | AudioParam | dict,
        **kwargs: Any,
    ) -> Response[list[float]]:
        """Create embedding for audio.

        Args:
            audio: Audio.

        Returns:
            Audio embedding.
        """
        ...

    @operation()
    def compare(
        self,
        embedding1: list[float],
        embedding2: list[float],
        dist_metric: str = "cosine",
        **kwargs,
    ) -> Response[float]:
        """Compare embeddings using cosine similarity.

        Args:
            embedding1: Embedding 1.
            embedding2: Embedding 2.

        Returns:
            Score defined by dist_metric.
        """
        ...

    @operation()
    def batch(
        self,
        audios: list[bytes | Audio | Audio | dict],
        **kwargs: Any,
    ) -> Response[list[list[float]]]:
        """Create embedding for batch of audios.

        Args:
            audios: List of audios.

        Returns:
            List of audio embeddings.
        """
        ...

    @operation()
    def embed_text(
        self,
        text: str,
        **kwargs: Any,
    ) -> Response[list[float]]:
        """Create embedding for text.

        Args:
            text: Text.

        Returns:
            Text embedding.
        """
        ...

    @operation()
    def batch_text(
        self,
        texts: list[str],
        **kwargs: Any,
    ) -> Response[list[list[float]]]:
        """Create embedding for batch of texts.

        Args:
            texts: List of texts.

        Returns:
            List of text embeddings.
        """
        ...

    @operation()
    def batch_audio_text(
        self,
        audios: list[bytes | Audio | AudioParam | dict],
        texts: list[str],
        **kwargs: Any,
    ) -> Response[list[list[float]]]:
        """Create embedding for batch of audio and text.

        Args:
            audios: List of audios.
            texts: List of texts.

        Returns:
            List of audio-text embeddings.
        """
        ...

    @operation()
    async def aembed(
        self,
        audio: bytes | Audio | AudioParam | dict,
        **kwargs: Any,
    ) -> Response[list[float]]:
        """Create embedding for audio.

        Args:
            audio: Audio.

        Returns:
            Audio embedding.
        """
        ...

    @operation()
    async def acompare(
        self,
        embedding1: list[float],
        embedding2: list[float],
        dist_metric: str = "cosine",
        **kwargs,
    ) -> Response[float]:
        """Compare embeddings using cosine similarity.

        Args:
            embedding1: Embedding 1.
            embedding2: Embedding 2.

        Returns:
            Score defined by dist_metric.
        """
        ...

    @operation()
    async def abatch(
        self,
        audios: list[bytes | Audio | Audio | dict],
        **kwargs: Any,
    ) -> Response[list[list[float]]]:
        """Create embedding for batch of audios.

        Args:
            audios: List of audios.

        Returns:
            List of audio embeddings.
        """
        ...

    @operation()
    async def aembed_text(
        self,
        text: str,
        **kwargs: Any,
    ) -> Response[list[float]]:
        """Create embedding for text.

        Args:
            text: Text.

        Returns:
            Text embedding.
        """
        ...

    @operation()
    async def abatch_text(
        self,
        texts: list[str],
        **kwargs: Any,
    ) -> Response[list[list[float]]]:
        """Create embedding for batch of texts.

        Args:
            texts: List of texts.

        Returns:
            List of text embeddings.
        """
        ...

    @operation()
    async def abatch_audio_text(
        self,
        audios: list[bytes | Audio | AudioParam | dict],
        texts: list[str],
        **kwargs: Any,
    ) -> Response[list[list[float]]]:
        """Create embedding for batch of audio and text.

        Args:
            audios: List of audios.
            texts: List of texts.

        Returns:
            List of audio-text embeddings.
        """
        ...



import base64
from typing import Any

import httpx

from verse.content.audio import Audio
from verse.core import Provider, Response

from .._models import TextToSpeechResult


class Hyperbolic(Provider):
    base_url: str
    api_key: str

    def __init__(
        self,
        api_key: str = "",
        base_url: str = "https://api.hyperbolic.xyz/v1/audio/generation",
        **kwargs,
    ):
        self.base_url = base_url
        self.api_key = api_key

    def generate(
        self,
        text: str,
        speed: float | None = None,
        **kwargs: Any,
    ) -> Response[TextToSpeechResult]:
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.api_key}",
        }
        data = {"text": text, "speed": speed or 1}
        res = httpx.post(
            self.base_url, headers=headers, json=data, timeout=None
        ).json()
        return TextToSpeechResult(
            audio=Audio(data=base64.b64decode(res["audio"])),
        )



from verse.content.audio import Audio, AudioParam
from verse.core import OperationParser


class SpeechRecognitionOperationParser(OperationParser):
    def get_audio(self) -> Audio:
        audio = self.get_arg("audio")
        return SpeechRecognitionOperationParser.normalize_audio(audio)

    @staticmethod
    def normalize_audio(audio) -> Audio:
        if isinstance(audio, str):
            return Audio(path=audio)
        elif isinstance(audio, bytes):
            return Audio(data=audio)
        elif isinstance(audio, dict):
            return Audio(param=AudioParam.from_dict(audio))
        elif isinstance(audio, AudioParam):
            return Audio(param=audio)
        return audio



import base64
from typing import Any

import httpx

from verse.content.image import Image, ImageParam
from verse.core import Provider, Response

from .._models import ImageGenerationResult


class Hyperbolic(Provider):
    base_url: str
    api_key: str
    model: str

    def __init__(
        self,
        api_key: str = "",
        base_url: str = "https://api.hyperbolic.xyz/v1/image/generation",
        model: str = "SD2",
        **kwargs,
    ):
        self.base_url = base_url
        self.api_key = api_key
        self.model = model

    def generate(
        self,
        prompt: str,
        negative_prompt: str = "",
        reference_image: str | bytes | Image | ImageParam | dict | None = None,
        width: int = 1024,
        height: int = 1024,
        **kwargs: Any,
    ) -> Response[ImageGenerationResult]:
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.api_key}",
        }
        data = {
            "model_name": self.model,
            "prompt": prompt,
            "steps": kwargs.get("steps", 30),
            "cfg_scale": kwargs.get("cfg_scale", 5),
            "enable_refiner": kwargs.get("enable_refiner", False),
            "height": width,
            "width": height,
            "backend": "auto",
        }
        if negative_prompt:
            data["negative_prompt"] = negative_prompt
        if reference_image:
            data["enable_reference"] = True
            data["image"] = Image.load(reference_image).convert(type="base64")
        res = httpx.post(
            self.base_url, headers=headers, json=data, timeout=None
        ).json()
        try:
            images: list = []
            for image in res["images"]:
                images.append(Image(data=base64.b64decode(image["image"])))
            return Response(result=ImageGenerationResult(images=images))
        except Exception as e:
            print(res)
            raise e



"""
Image embedding using CLIP model.
"""

__all__ = ["Clip"]

from typing import Any

import torch
from transformers import CLIPModel, CLIPProcessor

from verse.core import Context, Operation, Provider, Response

from .._operation import ImageEmbeddingOperation
from .._operation_parser import ImageEmbeddingOperationParser


class Clip(Provider):
    model: str
    device: str | None

    _device: Any
    _model: Any
    _processor: Any

    def __init__(
        self,
        model: str = "openai/clip-vit-base-patch32",
        device: str | None = None,
        **kwargs,
    ):
        """Intialize.

        Args:
            model:
                CLIP model name.
            device:
                Pytorch device name.
        """
        self.model = model
        self.device = device

        self._model = None
        self._device = None
        self._processor = None

    def __setup__(self, context: Context | None = None) -> None:
        if self._model:
            return
        self._device = torch.device(
            "cuda" if torch.cuda.is_available() else "cpu"
        )
        if self.device is not None:
            self._device = self.device
        self._model = CLIPModel.from_pretrained(self.model).to(self._device)
        self._processor = CLIPProcessor.from_pretrained(self.model)

    def __run__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        self.__setup__(context=context)
        op_parser = ImageEmbeddingOperationParser(operation)
        # EMBED IMAGE
        if op_parser.op_equals(ImageEmbeddingOperation.EMBED):
            inputs = self._processor(
                images=op_parser.get_image_as_pil(),
                return_tensors="pt",
                padding=True,
            ).to(self._device)
            with torch.no_grad():
                embeddings = self._model.get_image_features(**inputs)
            embeddings = embeddings / embeddings.norm(dim=-1, keepdim=True)
            result = embeddings.cpu().numpy().flatten().tolist()
        # BATCH EMBED IMAGE
        elif op_parser.op_equals(ImageEmbeddingOperation.BATCH):
            inputs = self._processor(
                images=op_parser.get_images_as_pil(),
                return_tensors="pt",
                padding=True,
            ).to(self._device)
            with torch.no_grad():
                embeddings = self._model.get_image_features(**inputs)
            embeddings = embeddings / embeddings.norm(dim=-1, keepdim=True)
            result = [
                embedding.cpu().numpy().flatten().tolist()
                for embedding in embeddings
            ]
        # EMBED TEXT
        elif op_parser.op_equals(ImageEmbeddingOperation.EMBED_TEXT):
            inputs = self._processor(
                text=op_parser.get_text(),
                return_tensors="pt",
                padding=True,
            ).to(self._device)
            with torch.no_grad():
                embeddings = self._model.get_text_features(**inputs)
            embeddings = embeddings / embeddings.norm(dim=-1, keepdim=True)
            result = embeddings.cpu().numpy().flatten().tolist()
        # BATCH EMBED TEXT
        elif op_parser.op_equals(ImageEmbeddingOperation.BATCH_TEXT):
            inputs = self._processor(
                text=op_parser.get_texts(),
                return_tensors="pt",
                padding=True,
            ).to(self._device)
            with torch.no_grad():
                embeddings = self._model.get_text_features(**inputs)
            embeddings = embeddings / embeddings.norm(dim=-1, keepdim=True)
            result = [
                embedding.cpu().numpy().flatten().tolist()
                for embedding in embeddings
            ]
        # BATCH EMBED IMAGE TEXT
        elif op_parser.op_equals(ImageEmbeddingOperation.BATCH_IMAGE_TEXT):
            inputs = self._processor(
                text=op_parser.get_texts(),
                images=op_parser.get_images_as_pil(),
                return_tensors="pt",
                padding=True,
            ).to(self._device)
            with torch.no_grad():
                image_embeddings = self._model.get_image_features(
                    pixel_values=inputs["pixel_values"]
                )
                text_embeddings = self._model.get_text_features(
                    input_ids=inputs["input_ids"],
                    attention_mask=inputs["attention_mask"],
                )
            image_embeddings = image_embeddings / image_embeddings.norm(
                dim=-1, keepdim=True
            )
            text_embeddings = text_embeddings / text_embeddings.norm(
                dim=-1, keepdim=True
            )
            result = (
                [
                    embedding.cpu().numpy().flatten().tolist()
                    for embedding in image_embeddings
                ],
                [
                    embedding.cpu().numpy().flatten().tolist()
                    for embedding in text_embeddings
                ],
            )
        # COMPARE
        elif op_parser.op_equals(ImageEmbeddingOperation.COMPARE):
            embedding1 = torch.tensor(op_parser.get_embedding1())
            embedding2 = torch.tensor(op_parser.get_embedding2())
            similarity = torch.nn.functional.cosine_similarity(
                embedding1, embedding2, dim=0
            )
            result = similarity.item()
        else:
            return super().__run__(operation, context, **kwargs)
        return Response(result=result)



from verse.content.image import Image
from verse.core import OperationParser


class ImageEmbeddingOperationParser(OperationParser):
    def get_image(self) -> Image:
        image = self.get_arg("image")
        return Image.load(image)

    def get_image_as_pil(self):
        image = self.get_image()
        return image.convert("pil")

    def get_images(self) -> list[Image]:
        result = []
        images = self.get_arg("images")
        for image in images:
            result.append(Image.load(image))
        return result

    def get_images_as_pil(self):
        result = []
        images = self.get_arg("images")
        for image in images:
            result.append(Image.load(image).convert("pil"))
        return result

    def get_text(self) -> str:
        return self.get_arg("text")

    def get_texts(self) -> list[str]:
        return self.get_arg("texts")

    def get_embedding1(self) -> list[float]:
        return self.get_arg("embedding1")

    def get_embedding2(self) -> list[float]:
        return self.get_arg("embedding2")



"""
Speech recognition using Whisper.
"""

__all__ = ["Whisper"]

import functools
from typing import Any

import av
import numpy as np
import whisper
from whisper import Whisper as OpenAIWhisper

from verse.content.audio import Audio, AudioParam
from verse.core import Context, Provider, Response

from .._models import SpeechRecognitionResult
from .._operation_parser import SpeechRecognitionOperationParser

whisper.torch.load = functools.partial(whisper.torch.load, weights_only=True)


class Whisper(Provider):
    model: str
    device: str | None
    fp16: bool

    _model: OpenAIWhisper

    def __init__(
        self,
        model: str = "tiny",
        device: str | None = None,
        fp16: bool = False,
        **kwargs,
    ):
        """Intialize.

        Args:
            model:
                Whisper model name.
            device:
                Pytorch device name.
            fp16:
                fp16 option.
        """
        self.model = model
        self.device = device
        self.fp16 = fp16
        self._model = None

    def __setup__(self, context: Context | None = None) -> None:
        if self._model:
            return
        self._model = whisper.load_model(name=self.model, device=self.device)

    def transcribe(
        self,
        audio: str | bytes | Audio | AudioParam | dict,
        **kwargs: Any,
    ) -> Response[SpeechRecognitionResult]:
        naudio = SpeechRecognitionOperationParser.normalize_audio(audio)
        waudio = self._extract_and_resample_audio(
            naudio.convert(type="stream")
        )
        waudio = whisper.pad_or_trim(waudio)
        result = whisper.transcribe(self._model, waudio, fp16=self.fp16)
        result = SpeechRecognitionResult(
            text=result["text"], language=result["language"], verbose=result
        )
        return Response(result=result)

    def detect_language(
        self,
        audio: str | bytes | Audio | AudioParam | dict,
        **kwargs: Any,
    ) -> Response[SpeechRecognitionResult]:
        naudio = SpeechRecognitionOperationParser.normalize_audio(audio)
        waudio = self._extract_and_resample_audio(
            naudio.convert(type="stream")
        )
        waudio = whisper.pad_or_trim(waudio)
        mel = whisper.log_mel_spectrogram(waudio).to(self._model.device)
        _, probs = self._model.detect_language(mel)
        return Response(
            result=SpeechRecognitionResult(language=max(probs, key=probs.get))
        )

    def _extract_and_resample_audio(self, stream, target_rate=16000):
        container = av.open(stream)
        stream = container.streams.audio[0]

        resampler = av.audio.resampler.AudioResampler(
            format="s16", layout="mono", rate=target_rate
        )

        samples = []

        for frame in container.decode(stream):
            resampled_frames = resampler.resample(frame)
            for resampled_frame in resampled_frames:
                samples.append(resampled_frame.to_ndarray())

        audio_data = np.concatenate(samples, axis=1)
        if audio_data.ndim > 1:
            audio_data = audio_data.mean(axis=0)

        audio_data = audio_data / np.max(np.abs(audio_data))
        if audio_data.dtype != np.float32:
            audio_data = audio_data.astype(np.float32)

        return audio_data



from verse.core import DataModel


class Usage(DataModel):
    prompt_tokens: int
    """Number of tokens in the prompt."""

    total_tokens: int
    """Total number of tokens used by the request."""


class TextEmbeddingResult(DataModel):
    embeddings: list[list[float]]
    """List of embeddings."""

    model: str | None = None
    """The model used for the embedding."""

    usage: Usage | None = None
    """Usage statistics for the embedding request."""



"""
Audio embedding provider using CLAP model.
"""

__all__ = ["Clap"]

from typing import Any

import numpy as np
import torch
from transformers import AutoFeatureExtractor, AutoProcessor, ClapModel

from verse.content.audio import Audio
from verse.core import Context, Operation, Provider, Response

from .._operation import AudioEmbeddingOperation
from .._operation_parser import AudioEmbeddingOperationParser


class Clap(Provider):
    model: str
    device: str | None

    _device: Any
    _model: Any
    _processor: Any

    def __init__(
        self,
        model: str = "laion/clap-htsat-fused",
        device: str | None = None,
        **kwargs,
    ):
        """Intialize.

        Args:
            model:
                CLAP model name.
            device:
                Pytorch device name.
        """
        self.model = model
        self.device = device

        self._model = None
        self._device = None
        self._processor = None

    def __setup__(self, context: Context | None = None) -> None:
        if self._model:
            return
        self._device = torch.device(
            "cuda" if torch.cuda.is_available() else "cpu"
        )
        if self.device is not None:
            self._device = self.device
        self._model = ClapModel.from_pretrained(self.model).to(self._device)
        self._processor = AutoProcessor.from_pretrained(self.model)

    def __run__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Response:
        self.__setup__()
        op_parser = AudioEmbeddingOperationParser(operation)

        if op_parser.op_equals(AudioEmbeddingOperation.EMBED):
            result = self._embed(audio=op_parser.get_audio())
        elif op_parser.op_equals(AudioEmbeddingOperation.EMBED_TEXT):
            result = self._embed_text(text=op_parser.get_text())
        elif op_parser.op_equals(AudioEmbeddingOperation.BATCH):
            result = self._batch(audios=op_parser.get_audios())
        elif op_parser.op_equals(AudioEmbeddingOperation.BATCH_TEXT):
            result = self._batch_text(texts=op_parser.get_texts())
        elif op_parser.op_equals(AudioEmbeddingOperation.BATCH_AUDIO_TEXT):
            result = self._batch_audio_text(
                audios=op_parser.get_audios(), texts=op_parser.get_texts()
            )
        elif op_parser.op_equals(AudioEmbeddingOperation.COMPARE):
            result = self._compare(
                embedding1=torch.Tensor(op_parser.get_embedding1()),
                embedding2=torch.Tensor(op_parser.get_embedding2()),
                dist_metric=op_parser.get_dist_metric(),
            )
        else:
            return super().__run__(operation, **kwargs)

        return Response(result=result)

    def _embed(
        self,
        audio: Audio,
        **kwargs: Any,
    ):
        audio = self._preprocess_audio(audio)

        inputs = self._processor(
            audios=audio,
            return_tensors="pt",
            padding=True,
            sampling_rate=48000,
        ).to(self._device)

        with torch.no_grad():
            outputs = self._model.get_audio_features(**inputs)

        result = outputs.cpu().numpy().flatten().tolist()
        return result

    def _embed_text(
        self,
        text: str,
        **kwargs: Any,
    ):
        inputs = self._processor(
            text=text, return_tensors="pt", padding=True
        ).to(self._device)

        with torch.no_grad():
            outputs = self._model.get_text_features(**inputs)

        result = outputs.cpu().numpy().flatten().tolist()
        return result

    def _batch(
        self,
        audios: list[Audio],
        **kwargs: Any,
    ):
        audios = [self._preprocess_audio(audio) for audio in audios]

        inputs = self._processor(
            audios=audios,
            return_tensors="pt",
            padding=True,
            sampling_rate=48000,
        ).to(self._device)

        with torch.no_grad():
            outputs = self._model.get_audio_features(**inputs)

        result = outputs.cpu().numpy().tolist()
        return result

    def _batch_text(
        self,
        texts: list[str],
        **kwargs: Any,
    ):
        inputs = self._processor(
            text=texts, return_tensors="pt", padding=True
        ).to(self._device)

        with torch.no_grad():
            outputs = self._model.get_text_features(**inputs)

        result = outputs.cpu().numpy().tolist()
        return result

    def _batch_audio_text(
        self,
        audios: list[Audio],
        texts: list[str],
        **kwargs: Any,
    ):
        audios = [self._preprocess_audio(audio) for audio in audios]

        inputs = self._processor(
            text=texts,
            audios=audios,
            return_tensors="pt",
            padding=True,
            sampling_rate=48000,
        ).to(self._device)

        with torch.no_grad():
            outputs = self._model(**inputs)

        result = [
            outputs.audio_embedding.cpu().numpy().tolist(),
            outputs.text_embedding.cpu().numpy().tolist(),
        ]
        return result

    def _compare(
        self,
        embedding1,
        embedding2,
        dist_metric: str = "cosine",
        **kwargs,
    ):
        """
        Compare embeddings using cosine similarity.
        embedding1: shape (n, d) where n is the number of embeddings and d is the dimension of the embeddings
        embedding2: shape (m, d) where m is the number of embeddings and d is the dimension of the embeddings
        dist_metric: distance metric to use. Currently only "cosine" is supported
        output: shape (n, m) where output[i, j] is the similarity between embedding1[i] and embedding2[j]
        """
        if dist_metric != "cosine":
            raise ValueError(f"Unsupported distance metric: {dist_metric}")
        if embedding1.shape[-1] != embedding2.shape[-1]:
            raise ValueError("Embeddings must have the same dimension")
        norm1 = torch.linalg.vector_norm(
            embedding1, ord=2, dim=-1, keepdim=True
        )
        norm2 = torch.linalg.vector_norm(
            embedding2, ord=2, dim=-1, keepdim=True
        )

        similarity = embedding1 @ embedding2.T / (norm1 * norm2.T)

        # TODO: add support for other distance metrics

        return similarity.cpu().numpy().tolist()

    def _preprocess_audio(self, audio: Audio):
        audio = audio.convert(
            type="bytes",
            format="wav",
            codec="pcm_s16le",
            rate=16000,
            channels=1,
        )
        array = np.frombuffer(audio, dtype=np.int16)
        array = array.astype(np.float32) / 32768.0

        return array



"""
Speech Recognition using Google Speech-to-Text.
"""

__all__ = ["GoogleSpeechToText"]

from typing import Any

from google.cloud import speech_v1p1beta1 as speech
from google.oauth2 import service_account

from verse.content.audio import Audio
from verse.core import Context, Operation, Provider, Response

from .._models import SpeechRecognitionResult
from .._operation import SpeechRecognitionOperation
from .._operation_parser import SpeechRecognitionOperationParser


class GoogleSpeechToText(Provider):
    def __init__(
        self,
        api_key_path: str = "",
        languages: list[str] = ["en-US", "en-GB", "es-ES", "fr-FR"],
        **kwargs,
    ):
        """Intialize.

        Args:
            api_key:
                Azure speech api key.
            service_region:
                Azure speech service region.
        """
        credentials = service_account.Credentials.from_service_account_file(
            api_key_path
        )
        self.client = speech.SpeechClient(credentials=credentials)
        self.languages = languages
        self.config = speech.RecognitionConfig(
            encoding=speech.RecognitionConfig.AudioEncoding.ENCODING_UNSPECIFIED,
            sample_rate_hertz=16000,
            language_code="en-US",  # Fallback language
            alternative_language_codes=languages,  # Possible languages
        )

    def __run__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        self.__setup__(context=context)
        op_parser = SpeechRecognitionOperationParser(operation)
        result: Any = None
        # TRANSCRIBE
        if op_parser.op_equals(SpeechRecognitionOperation.TRANSCRIBE):
            result = self._transcribe(audio=op_parser.get_audio())
        # DETECT_LANGUAGE
        elif op_parser.op_equals(SpeechRecognitionOperation.DETECT_LANGUAGE):
            result = self._detect_language(audio=op_parser.get_audio())
        else:
            return super().__run__(operation)
        return Response(result=result)

    def _transcribe(self, audio: Audio):
        audio = audio.convert(
            type="bytes",
            format="wav",
            codec="pcm_s16le",
            rate=16000,
            channels=1,
        )
        audio = speech.RecognitionAudio(content=audio)

        response = self.client.recognize(config=self.config, audio=audio)

        if len(response.results) == 0:
            return SpeechRecognitionResult(
                text="", verbose=response.results, language=None
            )

        # print("raw response:", response.results)
        return SpeechRecognitionResult(
            text=response.results[0].alternatives[0].transcript,
            verbose=response.results,
            language=response.results[0].language_code,
        )

    def _detect_language(self, audio: Audio):
        return self._transcribe(audio).language



# https://github.com/yl4579/StyleTTS2/blob/main/models.py
from .istftnet import AdaIN1d, Decoder
from munch import Munch
from pathlib import Path
from .plbert import load_plbert
from torch.nn.utils import weight_norm, spectral_norm
import json
import numpy as np
import os
import os.path as osp
import torch
import torch.nn as nn
import torch.nn.functional as F

class LinearNorm(torch.nn.Module):
    def __init__(self, in_dim, out_dim, bias=True, w_init_gain='linear'):
        super(LinearNorm, self).__init__()
        self.linear_layer = torch.nn.Linear(in_dim, out_dim, bias=bias)

        torch.nn.init.xavier_uniform_(
            self.linear_layer.weight,
            gain=torch.nn.init.calculate_gain(w_init_gain))

    def forward(self, x):
        return self.linear_layer(x)

class LayerNorm(nn.Module):
    def __init__(self, channels, eps=1e-5):
        super().__init__()
        self.channels = channels
        self.eps = eps

        self.gamma = nn.Parameter(torch.ones(channels))
        self.beta = nn.Parameter(torch.zeros(channels))

    def forward(self, x):
        x = x.transpose(1, -1)
        x = F.layer_norm(x, (self.channels,), self.gamma, self.beta, self.eps)
        return x.transpose(1, -1)
    
class TextEncoder(nn.Module):
    def __init__(self, channels, kernel_size, depth, n_symbols, actv=nn.LeakyReLU(0.2)):
        super().__init__()
        self.embedding = nn.Embedding(n_symbols, channels)

        padding = (kernel_size - 1) // 2
        self.cnn = nn.ModuleList()
        for _ in range(depth):
            self.cnn.append(nn.Sequential(
                weight_norm(nn.Conv1d(channels, channels, kernel_size=kernel_size, padding=padding)),
                LayerNorm(channels),
                actv,
                nn.Dropout(0.2),
            ))
        # self.cnn = nn.Sequential(*self.cnn)

        self.lstm = nn.LSTM(channels, channels//2, 1, batch_first=True, bidirectional=True)

    def forward(self, x, input_lengths, m):
        x = self.embedding(x)  # [B, T, emb]
        x = x.transpose(1, 2)  # [B, emb, T]
        m = m.to(input_lengths.device).unsqueeze(1)
        x.masked_fill_(m, 0.0)
        
        for c in self.cnn:
            x = c(x)
            x.masked_fill_(m, 0.0)
            
        x = x.transpose(1, 2)  # [B, T, chn]

        input_lengths = input_lengths.cpu().numpy()
        x = nn.utils.rnn.pack_padded_sequence(
            x, input_lengths, batch_first=True, enforce_sorted=False)

        self.lstm.flatten_parameters()
        x, _ = self.lstm(x)
        x, _ = nn.utils.rnn.pad_packed_sequence(
            x, batch_first=True)
                
        x = x.transpose(-1, -2)
        x_pad = torch.zeros([x.shape[0], x.shape[1], m.shape[-1]])

        x_pad[:, :, :x.shape[-1]] = x
        x = x_pad.to(x.device)
        
        x.masked_fill_(m, 0.0)
        
        return x

    def inference(self, x):
        x = self.embedding(x)
        x = x.transpose(1, 2)
        x = self.cnn(x)
        x = x.transpose(1, 2)
        self.lstm.flatten_parameters()
        x, _ = self.lstm(x)
        return x
    
    def length_to_mask(self, lengths):
        mask = torch.arange(lengths.max()).unsqueeze(0).expand(lengths.shape[0], -1).type_as(lengths)
        mask = torch.gt(mask+1, lengths.unsqueeze(1))
        return mask


class UpSample1d(nn.Module):
    def __init__(self, layer_type):
        super().__init__()
        self.layer_type = layer_type

    def forward(self, x):
        if self.layer_type == 'none':
            return x
        else:
            return F.interpolate(x, scale_factor=2, mode='nearest')

class AdainResBlk1d(nn.Module):
    def __init__(self, dim_in, dim_out, style_dim=64, actv=nn.LeakyReLU(0.2),
                 upsample='none', dropout_p=0.0):
        super().__init__()
        self.actv = actv
        self.upsample_type = upsample
        self.upsample = UpSample1d(upsample)
        self.learned_sc = dim_in != dim_out
        self._build_weights(dim_in, dim_out, style_dim)
        self.dropout = nn.Dropout(dropout_p)
        
        if upsample == 'none':
            self.pool = nn.Identity()
        else:
            self.pool = weight_norm(nn.ConvTranspose1d(dim_in, dim_in, kernel_size=3, stride=2, groups=dim_in, padding=1, output_padding=1))
        
        
    def _build_weights(self, dim_in, dim_out, style_dim):
        self.conv1 = weight_norm(nn.Conv1d(dim_in, dim_out, 3, 1, 1))
        self.conv2 = weight_norm(nn.Conv1d(dim_out, dim_out, 3, 1, 1))
        self.norm1 = AdaIN1d(style_dim, dim_in)
        self.norm2 = AdaIN1d(style_dim, dim_out)
        if self.learned_sc:
            self.conv1x1 = weight_norm(nn.Conv1d(dim_in, dim_out, 1, 1, 0, bias=False))

    def _shortcut(self, x):
        x = self.upsample(x)
        if self.learned_sc:
            x = self.conv1x1(x)
        return x

    def _residual(self, x, s):
        x = self.norm1(x, s)
        x = self.actv(x)
        x = self.pool(x)
        x = self.conv1(self.dropout(x))
        x = self.norm2(x, s)
        x = self.actv(x)
        x = self.conv2(self.dropout(x))
        return x

    def forward(self, x, s):
        out = self._residual(x, s)
        out = (out + self._shortcut(x)) / np.sqrt(2)
        return out
    
class AdaLayerNorm(nn.Module):
    def __init__(self, style_dim, channels, eps=1e-5):
        super().__init__()
        self.channels = channels
        self.eps = eps

        self.fc = nn.Linear(style_dim, channels*2)

    def forward(self, x, s):
        x = x.transpose(-1, -2)
        x = x.transpose(1, -1)
                
        h = self.fc(s)
        h = h.view(h.size(0), h.size(1), 1)
        gamma, beta = torch.chunk(h, chunks=2, dim=1)
        gamma, beta = gamma.transpose(1, -1), beta.transpose(1, -1)
        
        
        x = F.layer_norm(x, (self.channels,), eps=self.eps)
        x = (1 + gamma) * x + beta
        return x.transpose(1, -1).transpose(-1, -2)

class ProsodyPredictor(nn.Module):

    def __init__(self, style_dim, d_hid, nlayers, max_dur=50, dropout=0.1):
        super().__init__() 
        
        self.text_encoder = DurationEncoder(sty_dim=style_dim, 
                                            d_model=d_hid,
                                            nlayers=nlayers, 
                                            dropout=dropout)

        self.lstm = nn.LSTM(d_hid + style_dim, d_hid // 2, 1, batch_first=True, bidirectional=True)
        self.duration_proj = LinearNorm(d_hid, max_dur)
        
        self.shared = nn.LSTM(d_hid + style_dim, d_hid // 2, 1, batch_first=True, bidirectional=True)
        self.F0 = nn.ModuleList()
        self.F0.append(AdainResBlk1d(d_hid, d_hid, style_dim, dropout_p=dropout))
        self.F0.append(AdainResBlk1d(d_hid, d_hid // 2, style_dim, upsample=True, dropout_p=dropout))
        self.F0.append(AdainResBlk1d(d_hid // 2, d_hid // 2, style_dim, dropout_p=dropout))

        self.N = nn.ModuleList()
        self.N.append(AdainResBlk1d(d_hid, d_hid, style_dim, dropout_p=dropout))
        self.N.append(AdainResBlk1d(d_hid, d_hid // 2, style_dim, upsample=True, dropout_p=dropout))
        self.N.append(AdainResBlk1d(d_hid // 2, d_hid // 2, style_dim, dropout_p=dropout))
        
        self.F0_proj = nn.Conv1d(d_hid // 2, 1, 1, 1, 0)
        self.N_proj = nn.Conv1d(d_hid // 2, 1, 1, 1, 0)


    def forward(self, texts, style, text_lengths, alignment, m):
        d = self.text_encoder(texts, style, text_lengths, m)
        
        batch_size = d.shape[0]
        text_size = d.shape[1]
        
        # predict duration
        input_lengths = text_lengths.cpu().numpy()
        x = nn.utils.rnn.pack_padded_sequence(
            d, input_lengths, batch_first=True, enforce_sorted=False)
        
        m = m.to(text_lengths.device).unsqueeze(1)
        
        self.lstm.flatten_parameters()
        x, _ = self.lstm(x)
        x, _ = nn.utils.rnn.pad_packed_sequence(
            x, batch_first=True)
        
        x_pad = torch.zeros([x.shape[0], m.shape[-1], x.shape[-1]])

        x_pad[:, :x.shape[1], :] = x
        x = x_pad.to(x.device)
                
        duration = self.duration_proj(nn.functional.dropout(x, 0.5, training=self.training))
        
        en = (d.transpose(-1, -2) @ alignment)

        return duration.squeeze(-1), en
    
    def F0Ntrain(self, x, s):
        x, _ = self.shared(x.transpose(-1, -2))
        
        F0 = x.transpose(-1, -2)
        for block in self.F0:
            F0 = block(F0, s)
        F0 = self.F0_proj(F0)

        N = x.transpose(-1, -2)
        for block in self.N:
            N = block(N, s)
        N = self.N_proj(N)
        
        return F0.squeeze(1), N.squeeze(1)
    
    def length_to_mask(self, lengths):
        mask = torch.arange(lengths.max()).unsqueeze(0).expand(lengths.shape[0], -1).type_as(lengths)
        mask = torch.gt(mask+1, lengths.unsqueeze(1))
        return mask

class DurationEncoder(nn.Module):

    def __init__(self, sty_dim, d_model, nlayers, dropout=0.1):
        super().__init__()
        self.lstms = nn.ModuleList()
        for _ in range(nlayers):
            self.lstms.append(nn.LSTM(d_model + sty_dim, 
                                 d_model // 2, 
                                 num_layers=1, 
                                 batch_first=True, 
                                 bidirectional=True, 
                                 dropout=dropout))
            self.lstms.append(AdaLayerNorm(sty_dim, d_model))
        
        
        self.dropout = dropout
        self.d_model = d_model
        self.sty_dim = sty_dim

    def forward(self, x, style, text_lengths, m):
        masks = m.to(text_lengths.device)
        
        x = x.permute(2, 0, 1)
        s = style.expand(x.shape[0], x.shape[1], -1)
        x = torch.cat([x, s], axis=-1)
        x.masked_fill_(masks.unsqueeze(-1).transpose(0, 1), 0.0)
                
        x = x.transpose(0, 1)
        input_lengths = text_lengths.cpu().numpy()
        x = x.transpose(-1, -2)
        
        for block in self.lstms:
            if isinstance(block, AdaLayerNorm):
                x = block(x.transpose(-1, -2), style).transpose(-1, -2)
                x = torch.cat([x, s.permute(1, -1, 0)], axis=1)
                x.masked_fill_(masks.unsqueeze(-1).transpose(-1, -2), 0.0)
            else:
                x = x.transpose(-1, -2)
                x = nn.utils.rnn.pack_padded_sequence(
                    x, input_lengths, batch_first=True, enforce_sorted=False)
                block.flatten_parameters()
                x, _ = block(x)
                x, _ = nn.utils.rnn.pad_packed_sequence(
                    x, batch_first=True)
                x = F.dropout(x, p=self.dropout, training=self.training)
                x = x.transpose(-1, -2)
                
                x_pad = torch.zeros([x.shape[0], x.shape[1], m.shape[-1]])

                x_pad[:, :, :x.shape[-1]] = x
                x = x_pad.to(x.device)
        
        return x.transpose(-1, -2)
    
    def inference(self, x, style):
        x = self.embedding(x.transpose(-1, -2)) * np.sqrt(self.d_model)
        style = style.expand(x.shape[0], x.shape[1], -1)
        x = torch.cat([x, style], axis=-1)
        src = self.pos_encoder(x)
        output = self.transformer_encoder(src).transpose(0, 1)
        return output
    
    def length_to_mask(self, lengths):
        mask = torch.arange(lengths.max()).unsqueeze(0).expand(lengths.shape[0], -1).type_as(lengths)
        mask = torch.gt(mask+1, lengths.unsqueeze(1))
        return mask

# https://github.com/yl4579/StyleTTS2/blob/main/utils.py
def recursive_munch(d):
    if isinstance(d, dict):
        return Munch((k, recursive_munch(v)) for k, v in d.items())
    elif isinstance(d, list):
        return [recursive_munch(v) for v in d]
    else:
        return d

def build_model(path, device):
    config = Path(__file__).parent / 'config.json'
    assert config.exists(), f'Config path incorrect: config.json not found at {config}'
    with open(config, 'r') as r:
        args = recursive_munch(json.load(r))
    assert args.decoder.type == 'istftnet', f'Unknown decoder type: {args.decoder.type}'
    decoder = Decoder(dim_in=args.hidden_dim, style_dim=args.style_dim, dim_out=args.n_mels,
            resblock_kernel_sizes = args.decoder.resblock_kernel_sizes,
            upsample_rates = args.decoder.upsample_rates,
            upsample_initial_channel=args.decoder.upsample_initial_channel,
            resblock_dilation_sizes=args.decoder.resblock_dilation_sizes,
            upsample_kernel_sizes=args.decoder.upsample_kernel_sizes,
            gen_istft_n_fft=args.decoder.gen_istft_n_fft, gen_istft_hop_size=args.decoder.gen_istft_hop_size)
    text_encoder = TextEncoder(channels=args.hidden_dim, kernel_size=5, depth=args.n_layer, n_symbols=args.n_token)
    predictor = ProsodyPredictor(style_dim=args.style_dim, d_hid=args.hidden_dim, nlayers=args.n_layer, max_dur=args.max_dur, dropout=args.dropout)
    bert = load_plbert()
    bert_encoder = nn.Linear(bert.config.hidden_size, args.hidden_dim)
    for parent in [bert, bert_encoder, predictor, decoder, text_encoder]:
        for child in parent.children():
            if isinstance(child, nn.RNNBase):
                child.flatten_parameters()
    model = Munch(
        bert=bert.to(device).eval(),
        bert_encoder=bert_encoder.to(device).eval(),
        predictor=predictor.to(device).eval(),
        decoder=decoder.to(device).eval(),
        text_encoder=text_encoder.to(device).eval(),
    )
    for key, state_dict in torch.load(path, map_location='cpu', weights_only=True)['net'].items():
        assert key in model, key
        try:
            model[key].load_state_dict(state_dict)
        except:
            state_dict = {k[7:]: v for k, v in state_dict.items()}
            model[key].load_state_dict(state_dict, strict=False)
    return model



from typing import Any, Union

from verse.content.image import Image
from verse.core import Component, Response, operation

from ._models import VideoGenerationResult


class VideoGeneration(Component):
    @operation()
    def generate(
        self,
        prompt: str,
        key_frames: list[Union[Image, None]] = [],
        **kwargs: Any
    ) -> Response[VideoGenerationResult]:
        """
        Generate a video from a given text prompt.

        Args:
            prompt (str): The text prompt to generate the video from.
            key_frames (list[Image]): Optional list of key frames
                to guide the video generation.
                If empty, generate video using only the prompt.
                Defaults to an empty list.
            **kwargs (Any): Additional keyword arguments for video generation.

        Returns:
            VideoGenerationResult: The result of the video generation process.
        """
        ...

    @operation()
    async def agenerate(
        self,
        prompt: str,
        key_frames: list[Union[Image, None]] = [],
        **kwargs: Any
    ) -> Response[VideoGenerationResult]:
        """
        Asynchronously generate a video from a given text prompt.

        Args:
            prompt (str): The text prompt to generate the video from.
            key_frames (list[Image]): Optional list of key frames
                to guide the video generation.
                If empty, generate video using only the prompt.
                Defaults to an empty list.
            **kwargs (Any): Additional keyword arguments for video generation.

        Returns:
            VideoGenerationResult: The result of the video generation process.
        """
        ...



import asyncio
import base64
import time
from typing import Any, Optional, Union

import httpx

try:
    from runwayml import AsyncRunwayML, RunwayML
except:
    assert (
        False
    ), "runwayml package not found. Please install using 'pip install runwayml'"

from verse.content.image import Image
from verse.content.video import Video
from verse.core import Provider

from .._models import VideoGenerationResult


class Runwayml(Provider):
    def __init__(self, api_key: str = "", **kwargs):
        self.api_key = api_key

    def generate(
        self, prompt: str, key_frames: list[Union[Image, None]], **kwargs: Any
    ):
        client = RunwayML(api_key=self.api_key)
        assert (
            len(key_frames) == 1
        ), "RunwayAI only supports image to video generation."

        model = kwargs.get("model", "gen3a_turbo")
        duration = kwargs.get("duration", 5)
        ratio = kwargs.get("ratio", "1280:768")
        seed = kwargs.get("seed", 123)
        watermark = kwargs.get("watermark", False)

        prompt_image = self._get_prompt_image(key_frames[0])
        task = client.image_to_video.create(
            model=model,
            duration=duration,
            ratio=ratio,
            seed=seed,
            watermark=watermark,
            prompt_image=prompt_image,
            prompt_text=prompt,
        )
        task_id = task.id

        time.sleep(1)
        task = client.tasks.retrieve(task_id)
        while task.status not in ["SUCCEEDED", "FAILED"]:
            time.sleep(5)
            task = client.tasks.retrieve(task_id)

        video_url = task.output[0]
        for _ in range(3):
            response = httpx.get(video_url)
            if response.status_code == 200:
                break

        assert (
            response.status_code == 200
        ), f"Failed to download from {video_url} - status code: {response.status_code}"

        return VideoGenerationResult(
            video=Video(data=response.content, url=video_url)
        )

    async def agenerate(
        self, prompt: str, key_frames: list[Union[Image, None]], **kwargs: Any
    ):
        client = AsyncRunwayML(api_key=self.api_key)
        assert (
            len(key_frames) == 1
        ), "RunwayAI only supports image to video generation."

        model = kwargs.get("model", "gen3a_turbo")
        duration = kwargs.get("duration", 5)
        ratio = kwargs.get("ratio", "1280:768")
        seed = kwargs.get("seed", 123)
        watermark = kwargs.get("watermark", False)

        prompt_image = self._get_prompt_image(key_frames[0])
        task = await client.image_to_video.create(
            model=model,
            duration=duration,
            ratio=ratio,
            seed=seed,
            watermark=watermark,
            prompt_image=prompt_image,
            prompt_text=prompt,
        )
        task_id = task.id

        task = await client.tasks.retrieve(task_id)
        while task.status not in ["SUCCEEDED", "FAILED"]:
            await asyncio.sleep(5)
            task = await client.tasks.retrieve(task_id)

        video_url = task.output[0]
        for _ in range(3):
            async with httpx.AsyncClient() as httpx_client:
                response = await httpx_client.get(video_url)
                if response.status_code == 200:
                    break

        assert (
            response.status_code == 200
        ), f"Failed to download from {video_url} - status code: {response.status_code}"

        return VideoGenerationResult(
            video=Video(data=response.content, url=video_url),
        )

    def _get_prompt_image(self, image: Image) -> str:
        if image.url is not None:
            return image.url
        elif image.base64 is not None:
            return image.base64
        elif image.path is not None:
            with open(image.path, "rb") as f:
                return base64.b64encode(f.read()).decode("utf-8")
        else:
            assert (
                False
            ), "image should have one of url/base64/path properties."



from verse.content.video import Video
from verse.core import DataModel


class VideoGenerationResult(DataModel):
    """
    Represents the result of a video generation process.

    Attributes:
        video (Video): The generated video.
    """

    video: Video



from typing import Any

from verse.core import Component, Response, operation

from ._models import TextToSpeechResult


class TextToSpeech(Component):
    @operation()
    def generate(
        self,
        text: str,
        speed: float | None = None,
        **kwargs: Any,
    ) -> Response[TextToSpeechResult]:
        """
        Generate speech audio from the given text.

        Args:
            text: The text input to be converted into speech.
            speech: Speech of the speech. Normal is 1.

        Returns:
            Generated speech.
        """
        ...

    @operation()
    async def agenerate(
        self,
        text: str,
        speed: float | None = None,
        **kwargs: Any,
    ) -> Response[TextToSpeechResult]:
        """
        Generate speech audio from the given text.

        Args:
            prompt (str): The text input to be converted into speech.
            speech: Speech of the speech. Normal is 1.

        Returns:
            Generated speech.
        """
        ...



from typing import Any

from verse.core import DataModel


class SpeechRecognitionResult(DataModel):
    """Speech recognition result."""

    text: str | None = None
    language: str | None = None
    verbose: Any | None = None
    """Transcribed text."""



from typing import Any

from verse.content.image import Image, ImageParam
from verse.core import Component, Response, operation

from ._models import ImageGenerationResult


class ImageGeneration(Component):
    @operation()
    def generate(
        self,
        prompt: str,
        negative_prompt: str = "",
        reference_image: str | bytes | Image | ImageParam | dict | None = None,
        width: int = 1024,
        height: int = 1024,
        **kwargs: Any,
    ) -> Response[ImageGenerationResult]:
        """
        Generate image from the given prompt.

        Args:
            prompt:
                The text prompt to generate image.
            negative_prompt:
                Prompt on what not to include.
            reference_image:
                Reference image for generation.
            width:
                Width of the image.
            height:
                Height of the image.
            steps (int):
                The number of denoising steps or iterations the
                model uses to generate the image from the initial noise.
            cfg_scale (float):
                Controls the strength of the guidance applied to
                steer the generation process toward the given text prompt.
            enable_refiner (bool):
                Toggles the use of a "refiner" model,
                which is a secondary process designed to enhance the quality,
                details, or resolution of the initial generated image.

        Returns:
            Generated image.
        """
        ...

    @operation()
    async def agenerate(
        self,
        prompt: str,
        negative_prompt: str = "",
        reference_image: str | bytes | Image | ImageParam | dict | None = None,
        width: int = 1024,
        height: int = 1024,
        **kwargs: Any,
    ) -> Response[ImageGenerationResult]:
        """
        Generate image from the given prompt.

        Args:
            prompt:
                The text prompt to generate image.
            negative_prompt:
                Prompt on what not to include.
            reference_image:
                Reference image for generation.
            width:
                Width of the image.
            height:
                Height of the image.
            steps (int):
                The number of denoising steps or iterations the
                model uses to generate the image from the initial noise.
            cfg_scale (float):
                Controls the strength of the guidance applied to
                steer the generation process toward the given text prompt.
            enable_refiner (bool):
                Toggles the use of a "refiner" model,
                which is a secondary process designed to enhance the quality,
                details, or resolution of the initial generated image.

        Returns:
            Generated image.
        """
        ...



from __future__ import annotations

__all__ = ["OpenAI"]

from typing import Any

from verse.core import Context, NCall, Operation, Provider, Response

from .._models import TextEmbeddingResult, Usage
from .._operation import TextEmbeddingOperation
from .._operation_parser import TextEmbeddingOperationParser


class OpenAI(Provider):
    model: str | None
    api_key: str | None
    organization: str | None
    project: str | None
    base_url: str | None
    timeout: float | None
    service_tier: str | None
    dimension: int | None
    nparams: dict

    _client: Any
    _aclient: Any

    def __init__(
        self,
        model: str | None = "text-embedding-3-small",
        api_key: str | None = None,
        organization: str | None = None,
        project: str | None = None,
        base_url: str | None = None,
        timeout: float | None = None,
        service_tier: str | None = None,
        dimension: int | None = None,
        nparams: dict = dict(),
    ):
        """Initialize.

        Args:
            model:
                OpenAI model to use for text embedding.
            api_key:
                OpenAI API key.
            organization:
                OpenAI organization.
            project:
                OpenAI project.
            base_url:
                OpenAI base url.
            timeout:
                Timeout for client.
            service_tier:
                OpenAI service tier.
            dimension:
                Embedding dimension.
            nparams:
                Native params for OpenAI client.
        """
        self.model = model
        self.api_key = api_key
        self.organization = organization
        self.project = project
        self.base_url = base_url
        self.timeout = timeout
        self.service_tier = service_tier
        self.dimension = dimension
        self.nparams = nparams

        self._client = None
        self._aclient = None

    def __setup__(self, context: Context | None = None) -> None:
        if self._client is not None:
            return

        from openai import OpenAI

        self._client = OpenAI(**self._get_client_params())

    async def __asetup__(self, context: Context | None = None) -> None:
        if self._aclient is not None:
            return

        from openai import AsyncOpenAI

        self._aclient = AsyncOpenAI(**self._get_client_params())

    def _get_client_params(self) -> dict[str, Any]:
        args: dict = {}
        if self.api_key is not None:
            args["api_key"] = self.api_key
        if self.organization is not None:
            args["organization"] = self.organization
        if self.project is not None:
            args["project"] = self.project
        if self.base_url is not None:
            args["base_url"] = self.base_url
        if self.timeout is not None:
            args["timeout"] = self.timeout
        args = args | self.nparams
        return args

    def __run__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        self.__setup__(context=context)
        op_parser = TextEmbeddingOperationParser(operation)
        ncall, state = self._get_ncall(
            op_parser,
            self._client,
        )
        if ncall is None:
            return super().__run__(
                operation=operation,
                context=context,
                **kwargs,
            )
        nresult = ncall.invoke()
        result = self._convert_nresult(
            nresult,
            state,
            op_parser,
        )
        return Response(result=result, native=dict(call=ncall, result=nresult))

    async def __arun__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        await self.__asetup__(context=context)
        op_parser = TextEmbeddingOperationParser(operation)
        ncall, state = self._get_ncall(
            op_parser,
            self._aclient,
        )
        if ncall is None:
            return await super().__arun__(
                operation=operation,
                context=context,
                **kwargs,
            )
        nresult = await ncall.ainvoke()
        result = self._convert_nresult(
            nresult,
            state,
            op_parser,
        )
        return Response(result=result, native=dict(call=ncall, result=nresult))

    def _get_ncall(
        self,
        op_parser: TextEmbeddingOperationParser,
        client: Any,
    ) -> tuple[NCall | None, dict]:
        call = None
        state: dict = {}
        nargs = op_parser.get_nargs()
        op_converter = OperationConverter(
            self.model, self.dimension, self.service_tier
        )
        # EMBED
        if op_parser.op_equals(TextEmbeddingOperation.EMBED):
            args = op_converter.convert_embed(
                texts=op_parser.get_texts(),
            )
            call = NCall(
                client.embeddings.create,
                args,
                nargs,
            )
        return call, state

    def _convert_nresult(
        self,
        nresult: Any,
        state: dict,
        op_parser: TextEmbeddingOperationParser,
    ) -> Any:
        result: Any = None
        result_converter = ResultConverter()
        # EMBED
        if op_parser.op_equals(TextEmbeddingOperation.EMBED):
            result = result_converter.convert_embed_result(
                nresult,
            )
        return result


class ResultConverter:
    def __init__(self):
        pass

    def convert_embed_result(
        self,
        nresult: Any,
    ) -> TextEmbeddingResult:
        embeddings = [e.embedding for e in nresult.data]
        usage = None
        if nresult.usage:
            usage = Usage(
                prompt_tokens=nresult.usage.prompt_tokens,
                total_tokens=nresult.usage.total_tokens,
            )
        return TextEmbeddingResult(embeddings=embeddings, usage=usage)


class OperationConverter:
    model: str | None
    dimension: int | None
    service_tier: str | None

    def __init__(
        self,
        model: str | None,
        dimension: int | None,
        service_tier: str | None,
    ):
        self.model = model
        self.dimension = dimension
        self.service_tier = service_tier

    def convert_embed(
        self,
        texts: list[str],
        **kwargs,
    ) -> dict:
        args: dict = {"input": texts, "model": self.model}
        if self.dimension:
            args["dimensions"] = self.dimension
        return args



"""
Image captioning using BLIP model.
"""

__all__ = ["Blip"]

from typing import Any

import torch
from transformers import BlipForConditionalGeneration, BlipProcessor

from verse.core import Context, Operation, Provider, Response

from .._operation import ImageCaptioningOperation
from .._operation_parser import ImageCaptioningOperationParser


class Blip(Provider):
    model: str
    device: str | None

    _device: Any
    _model: Any
    _processor: Any

    def __init__(
        self,
        model: str = "Salesforce/blip-image-captioning-base",
        device: str | None = None,
        **kwargs,
    ):
        """Intialize.

        Args:
            model:
                BLIP model name.
            device:
                Pytorch device name.
        """
        self.model = model
        self.device = device

        self._model = None
        self._device = None
        self._processor = None

    def __setup__(self, context: Context | None = None) -> None:
        if self._model:
            return
        self._device = torch.device(
            "cuda" if torch.cuda.is_available() else "cpu"
        )
        if self.device is not None:
            self._device = self.device
        self._model = BlipForConditionalGeneration.from_pretrained(
            "Salesforce/blip-image-captioning-base"
        ).to(self._device)
        self._processor = BlipProcessor.from_pretrained(
            "Salesforce/blip-image-captioning-base"
        )

    def __run__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        self.__setup__(context=context)
        op_parser = ImageCaptioningOperationParser(operation)
        # CAPTION
        if op_parser.op_equals(ImageCaptioningOperation.CAPTION):
            inputs = self._processor(
                op_parser.get_image_as_pil(),
                return_tensors="pt",
            ).to(self._device)
            max_length = op_parser.get_max_length()
            with torch.no_grad():
                if max_length:
                    outputs = self._model.generate(
                        **inputs, max_length=max_length
                    )
                else:
                    outputs = self._model.generate(**inputs)
            caption = self._processor.decode(
                outputs[0], skip_special_tokens=True
            )
            result = caption
        # BATCH
        elif op_parser.op_equals(ImageCaptioningOperation.BATCH):
            inputs = self._processor(
                op_parser.get_images_as_pil(),
                return_tensors="pt",
            ).to(self._device)
            max_length = op_parser.get_max_length()
            with torch.no_grad():
                if max_length:
                    outputs = self._model.generate(
                        **inputs, max_length=max_length
                    )
                else:
                    outputs = self._model.generate(**inputs)
            captions = [
                self._processor.decode(output, skip_special_tokens=True)
                for output in outputs
            ]
            result = captions
        else:
            return super().__run__(
                operation,
                context,
                **kwargs,
            )
        return Response(result=result)



from verse.content.audio import Audio
from verse.core import DataModel


class TextToSpeechResult(DataModel):
    """
    TextToSpeechResult is a data model representing
    the result of a text-to-speech operation.
    """

    audio: Audio



from typing import Any

from verse.content.audio import Audio, AudioParam
from verse.core import Component, Response, operation

from ._models import SpeechRecognitionResult


class SpeechRecognition(Component):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @operation()
    def transcribe(
        self,
        audio: str | bytes | Audio | AudioParam | dict,
        **kwargs: Any,
    ) -> Response[SpeechRecognitionResult]:
        """Transcribe audio and convert to text.

        Args:
            audio: Audio to transcibe.

        Returns:
            Transciption result.
        """
        ...

    @operation()
    def detect_language(
        self,
        audio: str | bytes | Audio | AudioParam | dict,
        **kwargs: Any,
    ) -> Response[SpeechRecognitionResult]:
        """Detect langauge in audio.

        Args:
            audio: Audio to recognize.

        Returns:
            Detection result.
        """
        ...

    @operation()
    async def atranscribe(
        self,
        audio: str | bytes | Audio | AudioParam | dict,
        **kwargs: Any,
    ) -> Response[SpeechRecognitionResult]:
        """Transcribe audio and convert to text.

        Args:
            audio: Audio to transcibe.

        Returns:
            Transciption result.
        """
        ...

    @operation()
    async def adetect_language(
        self,
        audio: str | bytes | Audio | AudioParam | dict,
        **kwargs: Any,
    ) -> Response[SpeechRecognitionResult]:
        """Detect langauge in audio.

        Args:
            audio: Audio to recognize.

        Returns:
            Detection result.
        """
        ...



# https://github.com/yl4579/StyleTTS2/blob/main/Utils/PLBERT/util.py
from transformers import AlbertConfig, AlbertModel

class CustomAlbert(AlbertModel):
    def forward(self, *args, **kwargs):
        # Call the original forward method
        outputs = super().forward(*args, **kwargs)
        # Only return the last_hidden_state
        return outputs.last_hidden_state

def load_plbert():
    plbert_config = {'vocab_size': 178, 'hidden_size': 768, 'num_attention_heads': 12, 'intermediate_size': 2048, 'max_position_embeddings': 512, 'num_hidden_layers': 12, 'dropout': 0.1}
    albert_base_configuration = AlbertConfig(**plbert_config)
    bert = CustomAlbert(albert_base_configuration)
    return bert



import asyncio
import time
from typing import Any, Optional, Union

import httpx

try:
    from lumaai import NOT_GIVEN, AsyncLumaAI, LumaAI
except:
    assert False, "lumaai package not found. Please install using 'pip install lumaai'"

from verse.content.image import Image
from verse.content.video import Video
from verse.core import Provider

from .._models import VideoGenerationResult


class Lumaai(Provider):
    def __init__(
        self,
        api_key: str = "",
        **kwargs,
    ):
        self.api_key = api_key

    def generate(
        self, prompt: str, key_frames: list[Union[Image, None]], **kwargs: Any
    ):
        client = LumaAI(auth_token=self.api_key)

        key_frames_dict = self._get_keyframes(key_frames)
        if key_frames_dict is None:
            key_frames_dict = NOT_GIVEN

        aspect_ratio = self._get_aspect_ratio(**kwargs)
        generation = client.generations.create(
            prompt=prompt, aspect_ratio=aspect_ratio, keyframes=key_frames_dict
        )

        completed = False
        while not completed:
            generation = client.generations.get(id=generation.id)
            if generation.state == "completed":
                completed = True
            elif generation.state == "failed":
                raise RuntimeError(
                    f"Generation failed: {generation.failure_reason}"
                )

            time.sleep(3)

        # download the video
        video_url = generation.assets.video
        for _ in range(3):
            response = httpx.get(video_url)
            if response.status_code == 200:
                break

            time.sleep(3)

        assert (
            response.status_code == 200
        ), f"Failed to download from {video_url} - status code: {response.status_code}"

        return VideoGenerationResult(
            video=Video(data=response.content, url=video_url),
        )

    async def agenerate(
        self, prompt: str, key_frames: list[Union[Image, None]], **kwargs: Any
    ):
        client = AsyncLumaAI(auth_token=self.api_key)

        key_frames_dict = self._get_keyframes(key_frames)
        if key_frames_dict is None:
            key_frames_dict = NOT_GIVEN

        aspect_ratio = self._get_aspect_ratio(**kwargs)

        generation = await client.generations.create(
            prompt=prompt, aspect_ratio=aspect_ratio, keyframes=key_frames_dict
        )

        generation_id = generation.id
        completed = False
        while not completed:
            generation = await client.generations.get(id=generation_id)
            if generation.state == "completed":
                completed = True
            elif generation.state == "failed":
                raise RuntimeError(
                    f"Generation failed: {generation.failure_reason}"
                )

            await asyncio.sleep(3)

        # download the video
        video_url = generation.assets.video
        for _ in range(3):
            async with httpx.AsyncClient() as httpx_client:
                response = await httpx_client.get(video_url)
                if response.status_code == 200:
                    break

        assert (
            response.status_code == 200
        ), f"Failed to download from {video_url} - status code: {response.status_code}"

        return VideoGenerationResult(
            video=Video(data=response.content, url=video_url),
        )

    def _get_aspect_ratio(self, **kwargs: Any) -> str:
        aspect_ratio = kwargs.get("aspect_ratio", "16:9")
        assert aspect_ratio in [
            "1:1",
            "16:9",
            "9:16",
            "4:3",
            "3:4",
            "21:9",
            "9:21",
        ], 'aspect_ratio must be in \
            ["1:1", "16:9", "9:16", "4:3", "3:4", "21:9", "9:21"]'

        return aspect_ratio

    def _get_keyframes(
        self, key_frames: list[Union[Image, None]]
    ) -> Optional[dict]:
        assert (
            len(key_frames) <= 2
        ), "LumaAI only supports up to 2 key frames. \
            One as starting frame and One as ending frame."

        if len(key_frames) == 0:
            return None

        key_frames_dict = {}

        if key_frames[0] is not None:
            assert (
                key_frames[0].url != ""
            ), "key_frames[0].url must not be empty"

            key_frames_dict["frame0"] = {
                "type": "image",
                "url": key_frames[0].url,
            }

        if len(key_frames) == 2 and key_frames[1] is not None:
            assert (
                key_frames[1].url != ""
            ), "key_frames[1].url must not be empty"

            key_frames_dict["frame1"] = {
                "type": "image",
                "url": key_frames[1].url,
            }

        return key_frames_dict



from typing import Any

from verse.core import Component, Response, operation

from ._models import TextEmbeddingResult


class TextEmbedding(Component):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @operation()
    def embed(
        self,
        texts: str | list[str],
        **kwargs: Any,
    ) -> Response[TextEmbeddingResult]:
        """Create embedding for text.

        Args:
            texts: Texts to embed.

        Returns:
            Text embeddings.
        """
        ...

    @operation()
    def compare(
        self,
        embedding1: list[float],
        embedding2: list[float],
        **kwargs,
    ) -> Response[float]:
        """Compare embeddings using cosine similarity.

        Args:
            embedding1: Embedding 1.
            embedding2: Embedding 2.

        Returns:
            Similarity score.
        """
        ...

    @operation()
    async def aembed(
        self,
        texts: str | list[str],
        **kwargs: Any,
    ) -> Response[TextEmbeddingResult]:
        """Create embedding for text.

        Args:
            texts: Texts to embed.

        Returns:
            Text embeddings.
        """
        ...

    @operation()
    async def acompare(
        self,
        embedding1: list[float],
        embedding2: list[float],
        **kwargs,
    ) -> Response[float]:
        """Compare embeddings using cosine similarity.

        Args:
            embedding1: Embedding 1.
            embedding2: Embedding 2.

        Returns:
            Similarity score.
        """
        ...



from verse.content.image import Image, ImageParam
from verse.core import OperationParser


class ZeroShotImageClassificationOperationParser(OperationParser):
    def get_image(self) -> Image:
        image = self.get_arg("image")
        return self._normalize_image(image)

    def get_image_as_pil(self):
        image = self.get_image()
        return image.convert("pil")

    def get_images(self) -> list[Image]:
        result = []
        images = self.get_arg("images")
        for image in images:
            result.append(self._normalize_image(image))
        return result

    def get_images_as_pil(self):
        result = []
        images = self.get_arg("images")
        for image in images:
            result.append(self._normalize_image(image).convert("pil"))
        return result

    def _normalize_image(self, image) -> Image:
        if isinstance(image, str):
            return Image(path=image)
        elif isinstance(image, bytes):
            return Image(data=image)
        elif isinstance(image, dict):
            return Image(param=ImageParam.from_dict(image))
        elif isinstance(image, ImageParam):
            return Image(param=image)
        return image

    def get_classes(self) -> list[str]:
        return self.get_arg("classes")



import asyncio
import json
import logging
import time
from typing import Any, Union

import httpx

from verse.content.image import Image
from verse.content.video import Video
from verse.core import Provider

from .._models import VideoGenerationResult


class Hailuoai(Provider):
    def __init__(
        self,
        api_key: str = "",
        **kwargs,
    ):
        self.api_key = api_key

    def generate(
        self, prompt: str, key_frames: list[Union[Image, None]], **kwargs: Any
    ):
        model = kwargs.get("model", "video-01")

        task_id = self._invoke_video_generation(model, prompt)
        file_id = self._wait_for_task_completion(task_id)

        return self._fetch_video_result(file_id)

    async def agenerate(
        self, prompt: str, key_frames: list[Union[Image, None]], **kwargs: Any
    ):
        model = kwargs.get("model", "video-01")

        task_id = await self._ainvoke_video_generation(model, prompt)
        file_id = await self._await_for_task_completion(task_id)
        return await self._afetch_video_result(file_id)

    def _invoke_video_generation(self, model: str, prompt: str) -> str:
        url = "https://api.minimaxi.chat/v1/video_generation"
        payload = json.dumps({"prompt": prompt, "model": model})

        response = httpx.request(
            "POST",
            url,
            headers=self._get_headers(),
            data=payload,
        )

        assert (
            response.status_code == 200
        ), f"Failed to submit video generation - status code: {response.status_code}"

        response = response.json()

        status_code = response.get("base_resp", {}).get("status_code", 0)
        assert (
            status_code == 0
        ), f"Failed to invoke video generation - status code: {status_code}"

        return response["task_id"]

    def _wait_for_task_completion(self, task_id: str) -> str:
        url = (
            "https://api.minimaxi.chat/v1/query/video_generation?task_id="
            + task_id
        )

        while True:
            try:
                response = httpx.request(
                    "GET",
                    url,
                    headers=self._get_headers(),
                )
                status = response.json()["status"]
            except Exception as e:
                logging.error(f"Error getting status: {e}")
                time.sleep(5)
                continue

            if status == "Success":
                return response.json()["file_id"]

            if status == "Fail":
                raise Exception("Video generation failed")

            time.sleep(5)

    def _fetch_video_result(self, file_id: str):
        url = "https://api.minimaxi.chat/v1/files/retrieve?file_id=" + file_id
        response = httpx.request("GET", url, headers=self._get_headers())

        video_url = response.json()["file"]["download_url"]
        logging.info("Video download link: " + video_url)

        for _ in range(3):
            response = httpx.get(video_url)
            if response.status_code == 200:
                break

            time.sleep(3)

        return VideoGenerationResult(
            video=Video(data=response.content, url=video_url),
        )

    async def _ainvoke_video_generation(self, model: str, prompt: str) -> str:
        """
        Asynchronous function to invoke video generation.

        Args:
        - model (str): The video generation model to use.
        - prompt (str): The prompt to generate a video for.

        Returns:
        - task_id (str): The ID of the submitted video generation task.
        """
        url = "https://api.minimaxi.chat/v1/video_generation"
        payload = json.dumps({"prompt": prompt, "model": model})

        async with httpx.AsyncClient() as client:
            response = await client.post(
                url,
                headers=self._get_headers(),
                data=payload,
            )

            assert (
                response.status_code == 200
            ), f"Failed to submit video generation - status code: {response.status_code}"

            response = response.json()

            status_code = response.get("base_resp", {}).get("status_code", 0)
            assert (
                status_code == 0
            ), f"Failed to invoke video generation - status code: {status_code}"

            return response["task_id"]

    async def _await_for_task_completion(self, task_id: str) -> str:
        """
        Asynchronously waits for a video generation task to complete.

        Args:
        - task_id (str): ID of the task to wait for.

        Returns:
        - str: File ID of the generated video upon successful completion.

        Raises:
        - Exception: If video generation fails.
        """
        url = (
            "https://api.minimaxi.chat/v1/query/video_generation?task_id="
            + task_id
        )

        while True:
            try:
                async with httpx.AsyncClient() as client:
                    response = await client.get(
                        url,
                        headers=self._get_headers(),
                    )
                response.raise_for_status()  # Raises an exception for bad status codes
                status = response.json()["status"]
            except httpx.RequestError as e:
                logging.error(f"Error getting status: {e}")
                await asyncio.sleep(5)
                continue
            except Exception as e:
                logging.error(f"Error processing response: {e}")
                await asyncio.sleep(5)
                continue

            if status == "Success":
                return response.json()["file_id"]

            if status == "Fail":
                raise Exception("Video generation failed")

            await asyncio.sleep(5)

    async def _afetch_video_result(
        self, file_id: str
    ) -> VideoGenerationResult:
        """
        Async function to fetch video result.

        Args:
        - file_id (str): The ID of the file to retrieve.

        Returns:
        - VideoGenerationResult: The result of the video generation.
        """
        url = "https://api.minimaxi.chat/v1/files/retrieve?file_id=" + file_id

        async with httpx.AsyncClient() as client:
            response = await client.get(url, headers=self._get_headers())

        video_url = response.json()["file"]["download_url"]
        logging.info("Video download link: " + video_url)

        for _ in range(3):
            async with httpx.AsyncClient() as client:
                response = await client.get(video_url)
                if response.status_code == 200:
                    break
                await asyncio.sleep(3)

        return VideoGenerationResult(
            video=Video(data=response.content, url=video_url),
        )

    def _get_headers(self):
        return {
            "authorization": "Bearer " + self.api_key,
            "content-type": "application/json",
        }



from verse.content.audio import Audio, AudioParam
from verse.core import OperationParser


class AudioEmbeddingOperationParser(OperationParser):
    def get_audio(self) -> Audio:
        audio = self.get_arg("audio")
        return self._normalize_audio(audio)
    
    def get_audios(self) -> list[Audio]:
        result = []
        audios = self.get_arg("audios")
        for audio in audios:
            result.append(self._normalize_audio(audio))
        return result
    
    def _normalize_audio(self, audio) -> Audio:
        if isinstance(audio, str):
            return Audio(path=audio)
        elif isinstance(audio, bytes):
            return Audio(data=audio)
        elif isinstance(audio, dict):
            return Audio(param=AudioParam.from_dict(audio))
        elif isinstance(audio, AudioParam):
            return Audio(param=audio)
        return audio
    
    def get_text(self) -> str:
        return self.get_arg("text")

    def get_texts(self) -> list[str]:
        return self.get_arg("texts")

    def get_embedding1(self) -> list[float]:
        return self.get_arg("embedding1")

    def get_embedding2(self) -> list[float]:
        return self.get_arg("embedding2")
    
    def get_dist_metric(self) -> str:
        return self.get_arg("dist_metric")


import httpx
import io
import os
import torch
import soundfile as sf
from typing import Any
from verse.content.audio import Audio
from verse.core import Provider, Response

from .._models import TextToSpeechResult
from .kokoro.models import build_model
from .kokoro.kokoro import generate as kokoro_generate

class Default(Provider):
    MODEL_FILE_NAME = "kokoro-v0_19.pth"
    VOICE_NAMES = [
        'af', # Default voice is a 50-50 mix of Bella & Sarah
        'af_bella', 'af_sarah', 'am_adam', 'am_michael',
        'bf_emma', 'bf_isabella', 'bm_george', 'bm_lewis',
        'af_nicole', 'af_sky',
    ]
    
    def __init__(
        self
    ):
        try:
            self.device = 'cuda' if torch.cuda.is_available() else 'cpu'

            if not os.path.exists(self.MODEL_FILE_NAME):            
                url = "https://huggingface.co/hexgrad/Kokoro-82M/resolve/main/kokoro-v0_19.pth"
                print(f"Download model from {url}")
                response = httpx.get(url, timeout=600, verify=False, follow_redirects=True)
                assert response.status_code == 200, f"Failed to download model from {url} - status: {response.status_code}"
                with open(self.MODEL_FILE_NAME, "wb") as f:
                    f.write(response.content)

            self.model = build_model(self.MODEL_FILE_NAME, self.device)
            self.voices = {}
            for voice_name in self.VOICE_NAMES:
                voice_path = f"{voice_name}.pt"
                if not os.path.exists(voice_path):
                    # download voice if not existed
                    url = f"https://huggingface.co/hexgrad/Kokoro-82M/resolve/main/voices/{voice_name}.pt"
                    print(f"Download voice from {url}")
                    response = httpx.get(url, timeout=600, verify=False, follow_redirects=True)
                    with open(voice_path, "wb") as f:
                        f.write(response.content)

                self.voices[voice_name] = torch.load(voice_path, weights_only=True).to(self.device)
        except Exception as e:
            print("""If you have not installed Kokoro-TTS, please follow below instructions:
```
sudo apt-get -y install espeak-ng
pip install -q phonemizer torch transformers scipy munch 
```
                  
More info at: https://huggingface.co/hexgrad/Kokoro-82M
""")
            raise e

    def generate(
        self,
        text: str,
        speed: float | None = None,
        **kwargs: Any,
    ) -> Response[TextToSpeechResult]:
        voice = kwargs.get("voice", "af")
        assert voice in self.VOICE_NAMES, f"voice must be one of these: {self.VOICE_NAMES}"

        audio, _ = kokoro_generate(self.model, text, self.voices[voice], lang="a")
        
        audio_buffer = io.BytesIO()
        sf.write(audio_buffer, audio, 24000, format="wav")

        return TextToSpeechResult(
            audio=Audio(data=audio_buffer.getvalue()),
        )


from __future__ import annotations

__all__ = ["Cohere"]

from typing import Any

from verse.core import Context, NCall, Operation, Provider, Response

from .._models import TextEmbeddingResult, Usage
from .._operation import TextEmbeddingOperation
from .._operation_parser import TextEmbeddingOperationParser


class Cohere(Provider):
    api_key: str | None
    model: str | None
    input_type: str
    base_url: str | None
    client_name: str | None
    timeout: float | None
    nparams: dict

    _client: Any
    _aclient: Any

    def __init__(
        self,
        api_key: str | None = None,
        model: str = "embed-english-v3.0",
        input_type: str = "search_query",
        base_url: str | None = None,
        client_name: str | None = None,
        timout: float | None = None,
        nparams: dict = dict(),
    ):
        """Initialize.

        Args:
            api_key:
                Cohere API key.
            model:
                Cohere model.
            input_type:
                Cohere embedding input type.
            base_url:
                Cohere base url.
            client_name:
                Cohere client name.
            timout:
                Timeout for client.
            nparams:
                Native params for Cohere client.
        """
        self.api_key = api_key
        self.model = model
        self.input_type = input_type
        self.base_url = base_url
        self.client_name = client_name
        self.timeout = timout
        self.nparams = nparams

        self._client = None
        self._aclient = None

    def __setup__(self, context: Context | None = None) -> None:
        if self._client is not None:
            return

        from cohere import Client

        self._client = Client(**self.get_client_params())

    async def __asetup__(self, context: Context | None = None) -> None:
        if self._aclient is not None:
            return

        from cohere import AsyncClient

        self._aclient = AsyncClient(**self.get_client_params())

    def get_client_params(self) -> dict:
        args: dict = {}
        if self.api_key is not None:
            args["api_key"] = self.api_key
        if self.base_url is not None:
            args["base_url"] = self.base_url
        if self.client_name is not None:
            args["client_name"] = self.client_name
        if self.timeout is not None:
            args["timeout"] = self.timeout
        args = args | self.nparams
        return args

    def __run__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        self.__setup__(context=context)
        op_parser = TextEmbeddingOperationParser(operation)
        ncall, state = self._get_ncall(
            op_parser,
            self._client,
        )
        if ncall is None:
            return super().__run__(
                operation=operation, context=context, **kwargs
            )
        nresult = ncall.invoke()
        result = self._convert_nresult(
            nresult,
            state,
            op_parser,
        )
        return Response(result=result, native=dict(call=ncall, result=nresult))

    async def __arun__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        await self.__asetup__(context=context)
        op_parser = TextEmbeddingOperationParser(operation)
        ncall, state = self._get_ncall(
            op_parser,
            self._aclient,
        )
        if ncall is None:
            return await super().__arun__(
                operation=operation, context=context, **kwargs
            )
        nresult = await ncall.ainvoke()
        result = self._convert_nresult(
            nresult,
            state,
            op_parser,
        )
        return Response(result=result, native=dict(call=ncall, result=nresult))

    def _get_ncall(
        self,
        op_parser: TextEmbeddingOperationParser,
        client: Any,
    ) -> tuple[NCall | None, dict]:
        call = None
        state: dict = {}
        nargs = op_parser.get_nargs()
        op_converter = OperationConverter(
            self.model,
            self.input_type,
        )
        # EMBED
        if op_parser.op_equals(TextEmbeddingOperation.EMBED):
            args = op_converter.convert_embed(texts=op_parser.get_texts())
            call = NCall(
                client.embed,
                args,
                nargs,
            )
        return call, state

    def _convert_nresult(
        self,
        nresult: Any,
        state: dict,
        op_parser: TextEmbeddingOperationParser,
    ) -> Any:
        result: Any = None
        result_converter = ResultConverter()
        # COMPLETE
        if op_parser.op_equals(TextEmbeddingOperation.EMBED):
            result = result_converter.convert_embed_result(nresult)
        return result


class ResultConverter:
    _generation_id: str
    _model: str

    def __init__(self):
        pass

    def convert_embed_result(
        self,
        nresult: Any,
    ) -> TextEmbeddingResult:
        embeddings = [e for e in nresult.embeddings]
        usage = None
        if nresult.meta.billed_units:
            usage = Usage(
                prompt_tokens=nresult.meta.billed_units.input_tokens,
                total_tokens=nresult.meta.billed_units.input_tokens,
            )
        return TextEmbeddingResult(embeddings=embeddings, usage=usage)


class OperationConverter:
    model: str | None
    input_type: str

    def __init__(
        self,
        model: str | None,
        input_type: str,
    ):
        self.model = model
        self.input_type = input_type

    def convert_embed(
        self,
        texts: list[str],
        **kwargs,
    ) -> dict:
        args: dict = {
            "texts": texts,
            "model": self.model,
            "input_type": self.input_type,
        }
        return args



from verse.content.image import Image
from verse.core import DataModel


class ImageGenerationResult(DataModel):
    """
    The result of image generation operation.
    """

    images: list[Image]
